/* GLSL_multilight: different methods to display multiple lights, using plain GL, GLSL shaders or deferred shading
   Copyright (C) 2008 Angelo "Encelo" Theodorou
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <stdlib.h>

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glext.h> 

#include "init.h"
#include "textfile.h"


void InitLighting(struct Material *m)
{
	int i;

	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, m->ambient);
	glMaterialfv(GL_FRONT, GL_SPECULAR, m->specular);
	glMaterialf(GL_FRONT, GL_SHININESS, m->shininess);

	for(i = 0; i < NUM_LIGHTS; i++) {
		glLightf(GL_LIGHT0+i, GL_CONSTANT_ATTENUATION, 0.2f); //1.5f
		glLightf(GL_LIGHT0+i, GL_LINEAR_ATTENUATION, 0.1f); //0.5f
		glLightf(GL_LIGHT0+i, GL_QUADRATIC_ATTENUATION, 0.01f); //0.25f
	    glEnable(GL_LIGHT0+i);
	}
}


void InitFBOs(struct FBO *fbo)
{	
	glGenFramebuffersEXT(1, &(fbo->frame));
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo->frame);

	glGenRenderbuffersEXT(1, &(fbo->depth));
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, fbo->depth);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, fbo->width, fbo->height);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, fbo->depth);

	glGenTextures(3, fbo->texid);

	glBindTexture(GL_TEXTURE_2D, fbo->texid[0]);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F_ARB,  fbo->width, fbo->height, 0, GL_RGB, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, fbo->texid[0], 0);

	glBindTexture(GL_TEXTURE_2D, fbo->texid[1]);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F_ARB,  fbo->width, fbo->height, 0, GL_RGB, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_2D, fbo->texid[1], 0);

	glBindTexture(GL_TEXTURE_2D, fbo->texid[2]);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB16F_ARB,  fbo->width, fbo->height, 0, GL_RGB, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT2_EXT, GL_TEXTURE_2D, fbo->texid[2], 0);

	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status != GL_FRAMEBUFFER_COMPLETE_EXT) {
		printf("FBO Error: 0x%o\n", status);
		exit(-1);
	}
}


void DeleteFBOs(struct FBO *fbo)
{
	glDeleteTextures(3, fbo->texid);;
	glDeleteRenderbuffersEXT(1, &(fbo->depth));
	glDeleteFramebuffersEXT(1, &(fbo->frame));
}



void InitShaders(struct Shaders *sh)
{
	int i;
	char *vs, *fs, *filename;
	const char *vv, *ff;
	const char *vfiles[NUM_SHADERS] = {"pervertex1.vert", "perpixel1.vert", "pervertex3.vert", "perpixel3.vert", "deferredfirst.vert", "deferredsecond.vert"};
	const char *ffiles[NUM_SHADERS] = {"pervertex1.frag", "perpixel1.frag", "pervertex3.frag", "perpixel3.frag", "deferredfirst.frag", "deferredsecond.frag"};

	for (i = 0; i < NUM_SHADERS; i++) {

		sh->p[i] = glCreateProgram();

		filename = (char *)malloc(sizeof(char) * (strlen(SHADERS_DIR) + strlen(vfiles[i])));
		strcpy(filename, SHADERS_DIR);
		strcat(filename, vfiles[i]);

		sh->vert[i] = glCreateShader(GL_VERTEX_SHADER);
		vs = textFileRead(filename);
		printf("Vert[%d]: %s\n", i, filename);
		free(filename);
		vv = vs;
		glShaderSource(sh->vert[i], 1, &vv, NULL);
		free(vs);
		glCompileShader(sh->vert[i]);
		glAttachShader(sh->p[i], sh->vert[i]);

		filename = (char *)malloc(sizeof(char) * (strlen(SHADERS_DIR) + strlen(ffiles[i])));
		strcpy(filename, SHADERS_DIR);
		strcat(filename, ffiles[i]);

		sh->frag[i] = glCreateShader(GL_FRAGMENT_SHADER);
		fs = textFileRead(filename);
		printf("Frag[%d]: %s\n", i, filename);
		free(filename);
		ff = fs;
		glShaderSource(sh->frag[i], 1, &ff, NULL);
		free(fs);
		glCompileShader(sh->frag[i]);
		glAttachShader(sh->p[i], sh->frag[i]);

		glLinkProgram(sh->p[i]);
		printObjectInfoLog(sh->p[i]);
	}

	printf("\n");
}


void DeleteShaders(struct Shaders *sh)
{
	int i;

	for (i = 0; i < NUM_SHADERS; i++) {
		glDetachShader(sh->p[i], sh->frag[i]);
		glDeleteShader(sh->frag[i]);
		glDetachShader(sh->p[i], sh->vert[i]);
		glDeleteShader(sh->vert[i]);
		glDeleteProgram(sh->p[i]);
	}
}


GLuint LoadTexture(char *file)
{
	GLuint id;
	SDL_Surface *image;

	if (!(image = (SDL_Surface *)IMG_Load(file))) {
		printf("Texture %s load error", file);
		exit(-1);
	}

	glGenTextures(1, &id);
	glBindTexture(GL_TEXTURE_2D, id);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	if ((image->format)->Amask)
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, image->w, image->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, image->pixels);
	else
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, image->w, image->h, 0, GL_RGB, GL_UNSIGNED_BYTE, image->pixels);
	SDL_FreeSurface(image);

	return id;
}
